﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.Diagnostics;

namespace DynamixelCommander
{
    /* this class provides higher resolution timing
       needed for benchmarking */
    public class StopWatchMicrosecs : Stopwatch
    {
        public long ElapsedMicroseconds
        {
            get
            {
                // Stopwatch.Frequency returns the number of tocks per second; thus dividing ticks
                // per the frequency as a double, to keep precision, and then multiplying by 1000 000 gives you Microsecs
                return (long)(((double)base.ElapsedTicks / (double)Stopwatch.Frequency) * 1000000.0);
            }
        }
    }

    public class Dynamixel1BulkReadDeviceEntry
    {
        public byte dyn_id;
        public byte read_start_addr;
        public byte read_len;

        Dynamixel1BulkReadDeviceEntry() { }

        public Dynamixel1BulkReadDeviceEntry(byte b_id, byte b_read_startaddr, byte b_read_len)
        {
            this.dyn_id = b_id;
            this.read_start_addr = b_read_startaddr;
            this.read_len = b_read_len;
        }
    }

    public class Dynamixel1CommandGenerator
    {

        public const byte DYN1_INSTR_PING = 0x1;
        public const byte DYN1_INSTR_READ = 0x2;
        public const byte DYN1_INSTR_WRITE = 0x3;
        public const byte DYN1_INSTR_SYNC_WRITE = 0x83;
        public const byte DYN1_INSTR_BULK_READ = 0x92;
        public const byte DYN1_INSTR_WRITE_NO_REPLY = 0x33; // extension to DYN1 implemented in Seed Robotics fw
        public const byte DYN1_INSTR_REBOOT = 0x8; // Seed Robotics extension in DYN1; not part of DYN1 standard
        public const byte DYN1_INSTR_JUMP_TO_BOOTLDR = 0x9; // Seed Robotics extension in DYN1; not part of DYN1 standard

        public const byte ID_BROADCAST = 0xfe;


        public const byte DYN1_REPLY_SZ_READ_MIN = 6;
        public const byte DYN1_REPLY_SZ_WRITE = 6;
        public const byte DYN1_REPLY_SZ_PING = 6;

        public byte[] generate_dyn_packet(byte ID, byte INSTRUCTION, List<byte> PARAMS)
        {
            byte checksum = 0;

            List<byte> command_bytes = new List<byte>();

            command_bytes.Add(0xff);
            command_bytes.Add(0xff);
            command_bytes.Add(ID);
            command_bytes.Add((byte)(PARAMS.Count + 2));
            command_bytes.Add(INSTRUCTION);            
            command_bytes.AddRange(PARAMS);

            for (byte b = 2; b < command_bytes.Count; b++) {
                checksum = (byte) (checksum + command_bytes[b]);
            }
            checksum = (byte) (~checksum);

            command_bytes.Add(checksum);

            return command_bytes.ToArray();
        }

        public byte[] generate_bulkread_packet(List<Dynamixel1BulkReadDeviceEntry> device_entries)
        {
            List<byte> l_params = new List<byte>();

            l_params.Add(0x0);  // first param is always 0 for whatever reason
            foreach(Dynamixel1BulkReadDeviceEntry de in device_entries)
            {
                l_params.Add(de.read_len);
                l_params.Add(de.dyn_id);
                l_params.Add(de.read_start_addr);                
            }

            return generate_dyn_packet(ID_BROADCAST, DYN1_INSTR_BULK_READ, l_params);
        }


        public byte[] generate_write_packet(byte ID, byte wr_start_address, List<byte> data_to_write)
        {
            data_to_write.Insert(0, wr_start_address);

            return generate_dyn_packet(ID, DYN1_INSTR_WRITE, data_to_write);
        }

        public byte[] generate_read_packet(byte ID, byte start_addr, byte read_length)
        {
            List<byte> cmd_params = new List<byte>();
            cmd_params.Add(start_addr);
            cmd_params.Add(read_length);

            return generate_dyn_packet(ID, DYN1_INSTR_READ, cmd_params);
        }

        public byte[] generate_reboot_packet(byte ID)
        {
            return generate_dyn_packet(ID, DYN1_INSTR_REBOOT, new List<byte>());
        }

        public byte[] generate_ping_packet(byte ID)
        {
            return generate_dyn_packet(ID, DYN1_INSTR_PING, new List<byte>());
        }

        public byte[] generate_jumpt_to_bootldr_packet(byte ID, int delay_ms, byte baud_rate_register_low, byte baud_rate_register_high, int jump_address)
        {
            List<byte> cmd_params = new List<byte>();
            cmd_params.Add((byte)delay_ms);
            cmd_params.Add((byte)(delay_ms >> 8));
            cmd_params.Add(baud_rate_register_low);
            cmd_params.Add(baud_rate_register_high);
            cmd_params.Add((byte)jump_address);
            cmd_params.Add((byte)(jump_address >> 8));

            return generate_dyn_packet(ID, DYN1_INSTR_JUMP_TO_BOOTLDR, cmd_params);
        }

        public int get_word_value_from_reply(byte[] reply)
        {
            byte[] params_ = get_param_array_from_reply(reply);

            if (params_ != null && params_.Length == 2)
            {
                return (int)params_[0] + ((int)params_[0] * 256);
            }
            else {
                return -1;
            }
        }

        public int get_byte_value_from_reply(byte[] reply)
        {
            byte[] params_ = get_param_array_from_reply(reply);

            if (params_ != null && params_.Length == 1)
            {
                return (int)params_[0];
            }
            else {
                return -1;
            }
        }

        public byte[] get_param_array_from_reply(byte[] reply)
        {

            // locate the header in the reply;
            // do so to trim any data that could had been left in the buffer
            bool header_found = false;
            for(byte b=0; b < reply.Length - 1; b++)
            {
                if (reply[b] == 0xff && reply[b+1] == 0xff) // we can do b+1 without checkoing boundaries bc the for loop is b < reply.length-1
                {
                    if (b != 0) // trim the beginning of the array
                    {
                        reply = reply.Skip(b).ToArray();
                    }
                    header_found = true;
                    break;
                }
            }

            if (!header_found || reply.Length < 6) {
                return null;
            }

            // see if packet is complete
            int packet_len = (int)reply[3];

            if (reply.Length < packet_len + 4)
            {
                return null;
            }
                
            int nr_params = (int)reply[3] - 2;

            byte[] out_array = new byte[nr_params];
            Array.Copy(reply, 5, out_array, 0, nr_params);

            return out_array;
        }

        public byte[] get_dyn_reply(SerialPort port, byte expected_reply_size, long timeout_uSecs, ref int out_reply_size, ref long out_elapsed_usecs)
        {
            StopWatchMicrosecs timer = new StopWatchMicrosecs();

            timer.Start();

            while(timer.ElapsedMicroseconds < timeout_uSecs && port.BytesToRead < expected_reply_size)
            {
                ; // wait without thread.sleeping; we want maximum performance
            }
            timer.Stop();
            out_elapsed_usecs = timer.ElapsedMicroseconds;
            out_reply_size = port.BytesToRead; // cache size here as it may change while in execution and we need this size as reference


            if (out_elapsed_usecs > timeout_uSecs || out_reply_size < expected_reply_size) // timeout and/or not eneough bytes
            {
                Debug.Print("did not return properly: elapsed time {0}uSecs, bytes received {1}/expected {2}", out_elapsed_usecs, out_reply_size, expected_reply_size);
                return null;
            }

            byte[] received_bytes = new byte[expected_reply_size];
            port.Read(received_bytes, 0, expected_reply_size);

            return received_bytes;
        }

    }
}
