﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO.Ports;
using DynamixelCommander;


namespace SeedDynamixelBenchmarking
{
    class Program
    {

        private enum en_TestType { Invalid, Base, Extended };

        private const int NUMBER_OF_CYCLES_PER_BENCHMARK = 400;

        static int Main(string[] args)
        {
            string s_port = ""; long l_baud_bps=1000000; bool b_scan_ids=false; List<byte> l_dyn_ids_to_query = new List<byte>();
            byte b_ctb_read_start_addr=36; byte b_ctb_read_len=6; en_TestType e_test_type = en_TestType.Invalid; bool b_show_help = false;

            get_cmdline_params(args, ref s_port, ref l_baud_bps, ref b_scan_ids,  ref l_dyn_ids_to_query,
                               ref b_ctb_read_start_addr, ref b_ctb_read_len, ref e_test_type, ref b_show_help);

            if (b_show_help)
            {
                display_help();
                return 0;
            }

            // Ensure we have minum required arguments
            s_port = s_port.Trim();
            if (s_port.Length == 0)
            {
                Console.WriteLine("Serial port not specified. Use '-h' parameter for usage instructions.");
                return -1;
            }
            if (b_scan_ids == false && l_dyn_ids_to_query.Count == 0)
            {
                Console.WriteLine("No Dynamixel IDs specified; i.e. no -scanids or -dynid's listed. Use 'h' for instructions.");
                return -1;
            }
            if (e_test_type == en_TestType.Invalid)
            {
                Console.WriteLine("Invalid test type specified. Use 'h' for instructions.");
                return -1;
            }

            // actually open the serial port
            try
            {
                SerialPort c_port = new SerialPort(s_port, (int)l_baud_bps);
                c_port.RtsEnable = true; // use RTS transitions
                c_port.Open();

                // main program
                if (b_scan_ids)
                {
                    scan_bus_ids(ref c_port);
                }
                else
                {

                    Console.WriteLine("Running Benchmarks:");
                    Console.WriteLine("Will read {0} bytes from the {1} IDs, using different approaches:", b_ctb_read_len, l_dyn_ids_to_query.Count);
                    Console.WriteLine("=======================================================================");
                    Console.WriteLine("IDs will be queried in the following order: ");
                    foreach (byte b_id in l_dyn_ids_to_query) { Console.Write("[{0}] ", b_id); }
                    Console.WriteLine();
                    Console.WriteLine();


                    run_READ_benchmark(ref c_port, ref l_dyn_ids_to_query, b_ctb_read_start_addr, b_ctb_read_len);
                    run_BULK_READ_benchmark(ref c_port, ref l_dyn_ids_to_query, b_ctb_read_start_addr, b_ctb_read_len);

                    if (e_test_type == en_TestType.Extended) run_syncread_benchmarks(ref c_port, ref l_dyn_ids_to_query, b_ctb_read_start_addr, b_ctb_read_len);
                }

                c_port.Close();

                return 0;
            } catch (Exception ex)
            {
                Console.Write("> ERROR:{0}{1}", Environment.NewLine, ex.Message);
                return -2;
            }           
        }

        static void run_READ_benchmark(ref SerialPort port, ref List<byte> l_dyn_ids, byte b_read_startaddr, byte b_read_len)
        {
            Console.WriteLine();
            Console.WriteLine("#1 Individual READ commands:");
            Console.WriteLine("==========================================");
            Console.WriteLine();

            List<long> iteration_times = new List<long>(); int failed_runs = 0;
            long min_time_innercycle = long.MaxValue, max_time_innercycle = -1;
            byte min_time_innercycle_id = 0, max_time_innercycle_id = 0;
            long min_time_outercycle = long.MaxValue, max_time_outercycle = -1;

            Dynamixel1CommandGenerator dync = new Dynamixel1CommandGenerator();
            int bytes_received = 0; long reply_time_innercycle = 0;

            StopWatchMicrosecs outer_timer = new StopWatchMicrosecs(); bool run_failed=false; int cycle_nr = 0;
            for (cycle_nr = 0; cycle_nr < NUMBER_OF_CYCLES_PER_BENCHMARK; cycle_nr++)
            {
                outer_timer.Reset();
                outer_timer.Start();
                                
                foreach (byte b_id in l_dyn_ids)
                {
                    run_failed = false;

                    // make READ request
                    byte[] command = dync.generate_read_packet(b_id, b_read_startaddr, b_read_len);
                    port.Write(command, 0, command.Length);

                    byte[] reply = dync.get_dyn_reply(port, (byte) (Dynamixel1CommandGenerator.DYN1_REPLY_SZ_READ_MIN + b_read_len), 20000, ref bytes_received, ref reply_time_innercycle);
                    if (reply != null)
                    {
                        if (reply_time_innercycle < min_time_innercycle)
                        {
                            min_time_innercycle = reply_time_innercycle;
                            min_time_innercycle_id = b_id;
                        }

                        if (reply_time_innercycle > max_time_innercycle)
                        {
                            max_time_innercycle = reply_time_innercycle;
                            max_time_innercycle_id = b_id;
                        }
                    }
                    else
                    {
                        // run fails
                        run_failed = true;
                        port.ReadExisting(); // purge port
                        break;
                    }
                }
                outer_timer.Stop();

                if (!run_failed)
                {
                    min_time_outercycle = Math.Min(min_time_outercycle, outer_timer.ElapsedMicroseconds);
                    max_time_outercycle = Math.Max(max_time_outercycle, outer_timer.ElapsedMicroseconds);

                    iteration_times.Add(outer_timer.ElapsedMicroseconds);
                } else
                {
                    failed_runs++;
                }

                Console.Write("\rCycle #{0}: Cycle time {1}uSecs.  (Fastest so far {2}uSecs; slowest so far {3}uSecs. Failed runs: {4})     ", cycle_nr, (run_failed ? "FAILED" : outer_timer.ElapsedMicroseconds.ToString()), min_time_outercycle, max_time_outercycle, failed_runs );
                
            }


            // test cycle ended. run stats
            double avg, std_dev;
            if (iteration_times.Count > 1)
            {
                avg = iteration_times.Average();
                std_dev = Math.Sqrt(l_dyn_ids.Average(v => Math.Pow(v - avg, 2)));
            } else
            {
                avg = 0; std_dev = 0;
            }
            
            Console.WriteLine("\rStatistics:                                                                           ");
            Console.WriteLine("\tCycles ran         : {0,5}", cycle_nr);
            Console.WriteLine("\tCycles failed      : {0,5} {1}", failed_runs, (failed_runs > 0 ? "(check if the IDs are all connected and responding)" : ""));
            Console.WriteLine("\tFastest Cycle time : {0,5} uSec", min_time_outercycle);
            Console.WriteLine("\tSlowest Cycle time : {0,5} uSec", max_time_outercycle);
            Console.WriteLine("\tAverage Cycle time : {0,5} uSec", avg);
            Console.WriteLine("\tStandard Deviation : {0,5}", std_dev);
            Console.WriteLine("\t----------------------");
            Console.WriteLine("\tFastest Individual Device: ID {0}, reply time {1,5} uSec", min_time_innercycle_id, min_time_innercycle);
            Console.WriteLine("\tSlowest Individual Device: ID {0}, reply time {1,5} uSec", max_time_innercycle_id, max_time_innercycle);
            Console.WriteLine();
        }

        static void run_BULK_READ_benchmark(ref SerialPort port, ref List<byte> l_dyn_ids, byte b_read_startaddr, byte b_read_len)
        {
            Console.WriteLine("#2 Read multiple devices using one BULK_READ command:");
            Console.WriteLine("==========================================");
            Console.WriteLine();

            List<long> iteration_times = new List<long>(); int failed_runs = 0;
            long min_time_innercycle = long.MaxValue, max_time_innercycle = -1;
            byte min_time_innercycle_id = 0, max_time_innercycle_id = 0;
            long min_time_outercycle = long.MaxValue, max_time_outercycle = -1;

            Dynamixel1CommandGenerator dync = new Dynamixel1CommandGenerator();
            int bytes_received = 0; long inner_cycle_time = 0;

            StopWatchMicrosecs outer_timer = new StopWatchMicrosecs(); bool run_failed = false; int cycle_nr = 0;
            for (cycle_nr = 0; cycle_nr < NUMBER_OF_CYCLES_PER_BENCHMARK; cycle_nr++)
            {
                outer_timer.Reset();
                outer_timer.Start();

                // build BULK_READ command

                // create per device parameter triplets
                List<Dynamixel1BulkReadDeviceEntry> l_devicep = new List<Dynamixel1BulkReadDeviceEntry>();
                foreach(byte b_id in l_dyn_ids)
                {
                    l_devicep.Add(new Dynamixel1BulkReadDeviceEntry(b_id, b_read_startaddr, b_read_len));
                }

                byte[] command = dync.generate_bulkread_packet(l_devicep);
                port.Write(command, 0, command.Length);

                // get replies
                foreach (byte b_id in l_dyn_ids)
                {
                    run_failed = false;

                    byte[] reply = dync.get_dyn_reply(port, (byte)(Dynamixel1CommandGenerator.DYN1_REPLY_SZ_READ_MIN + b_read_len), 20000, ref bytes_received, ref inner_cycle_time);
                    if (reply != null)
                    {
                        if (inner_cycle_time < min_time_innercycle)
                        {
                            min_time_innercycle = inner_cycle_time;
                            min_time_innercycle_id = b_id;
                        }

                        if (inner_cycle_time > max_time_innercycle)
                        {
                            max_time_innercycle = inner_cycle_time;
                            max_time_innercycle_id = b_id;
                        }
                    }
                    else
                    {
                        // run fails
                        run_failed = true;
                        port.ReadExisting(); // purge port
                        break;
                    }
                }
                outer_timer.Stop();

                if (!run_failed)
                {
                    min_time_outercycle = Math.Min(min_time_outercycle, outer_timer.ElapsedMicroseconds);
                    max_time_outercycle = Math.Max(max_time_outercycle, outer_timer.ElapsedMicroseconds);

                    iteration_times.Add(outer_timer.ElapsedMicroseconds);
                }
                else
                {
                    failed_runs++;
                }

                Console.Write("\rCycle #{0}: Cycle time {1} uSec.  (Fastest so far {2} uSec; slowest so far {3} uSec. Failed runs: {4})     ", cycle_nr, (run_failed ? "FAILED" : outer_timer.ElapsedMicroseconds.ToString()), min_time_outercycle, max_time_outercycle, failed_runs);
            }


            // test cycle ended. run stats
            double avg, std_dev;
            if (iteration_times.Count > 1)
            {
                avg = iteration_times.Average();
                std_dev = Math.Sqrt(l_dyn_ids.Average(v => Math.Pow(v - avg, 2)));
            }
            else
            {
                avg = 0; std_dev = 0;
            }


            Console.WriteLine("\rStatistics:                                                                              ");
            Console.WriteLine("\tCycles ran         : {0,5}", cycle_nr);
            Console.WriteLine("\tCycles failed      : {0,5} {1}", failed_runs, (failed_runs > 0 ? "(check if the IDs are all connected and responding)" : ""));
            Console.WriteLine("\tFastest Cycle time : {0,5} uSec", min_time_outercycle);
            Console.WriteLine("\tSlowest Cycle time : {0,5} uSec", max_time_outercycle);
            Console.WriteLine("\tAverage Cycle time : {0,5} uSec", avg);
            Console.WriteLine("\tStandard Deviation : {0,5}", std_dev);
            Console.WriteLine("\t----------------------");
            Console.WriteLine("\tFastest Individual Device: ID {0}, reply time {1,5} uSec", min_time_innercycle_id, min_time_innercycle);
            Console.WriteLine("\tSlowest Individual Device: ID {0}, reply time {1,5} uSec", max_time_innercycle_id, max_time_innercycle);
            Console.WriteLine();
        }



        static void run_syncread_benchmarks(ref SerialPort port, ref List<byte> l_dyn_ids, byte b_read_startaddr, byte b_read_len)
        {


        }

        static void scan_bus_ids(ref SerialPort port)
        {
            Console.WriteLine();
            Console.WriteLine("Scanning Bus for Device IDs");
            Console.WriteLine("on serial port: {0} at {1} bps", port.PortName, port.BaudRate);
            Console.WriteLine("============================================");
            Console.WriteLine(">Only the scan will be performed; if you specified other actions, they will be ignored.");
            Console.WriteLine(">To run an actual benchmark test, explicitly list the Dynamixel IDs to use via the -dynids parameter.");
            Console.WriteLine();

            Dynamixel1CommandGenerator dync = new Dynamixel1CommandGenerator();

            int bytes_received = 0; long reply_time_usecs = 0, reply_time_ticks = 0;
            for (byte b=0; b < Dynamixel1CommandGenerator.ID_BROADCAST; b++ )
            {
                Console.Write("\rScanning ID {0} ...", b);

                byte[] dyn_command = dync.generate_ping_packet(b);

                port.Write(dyn_command, 0, dyn_command.Length);

                byte[] reply = dync.get_dyn_reply(port, Dynamixel1CommandGenerator.DYN1_REPLY_SZ_PING, 20000, ref bytes_received, ref reply_time_usecs);
                if (reply != null)
                {
                    Console.WriteLine("\rFound ID: {0}, replied in {1,5} uSecs)", b, reply_time_usecs, reply_time_ticks);
                }
            }

            Console.WriteLine("\rDone                                      ");
        }


        static void display_help()
        {
            Console.WriteLine();
            Console.WriteLine("Seed Robotics: Dynamixel Benchmarking tool");
            Console.WriteLine("==============================================");
            Console.WriteLine("Command line options:");
            Console.WriteLine("\t-port=[port name] - serial port to use");
            Console.WriteLine("\t-baud=[baud in bps] - baud rate to use; default=1000000/1Mbps");
            Console.WriteLine("\t-scanids - scans the bus for IDs and exits");
            Console.WriteLine("\t-dynids    =[id1,[id2],...] - list of IDs to use for the test, separed by semi-colon");
            Console.WriteLine("\t-ctbaddr   =[start address] - Control table address to start reading params; default=36");
            Console.WriteLine("\t-ctbreadlen=[read length] - Number of parameters to read from control table; default=6");
            Console.WriteLine("\t-test      =[base/extented] - type of benchmark tests to execute");
            Console.WriteLine("\t             base:     READ and BULK_READ benchmarking");
            Console.WriteLine("\t             extended: base + SYNC_READ (Richard iBotson's implementation, for Dynamixel 1)");
            Console.WriteLine("\t-h - this help screen");
        }

        static void get_cmdline_params(string[] args, ref string port, ref long baud_bps, ref bool scan_ids, ref List<byte> dyn_ids_to_query,
                                ref byte ctb_read_start_addr, ref byte ctb_read_len, ref en_TestType test_type, ref bool show_help)
        {
            test_type = en_TestType.Base; // initialize defaults; other default would have been initialized before being passed ByRef

            foreach (string arg in args)
            {
                string[] arg_params = arg.ToLower().Split('=');

                if (arg_params.Length == 1)
                {
                    switch (arg_params[0])
                    {
                        case "-h":
                        case "-?":
                            show_help = true;
                            break;

                        case "-scanids":
                            scan_ids = true;
                            break;

                        default:
                            Console.WriteLine("Invalid parameter: {0}", arg_params[0]);
                            show_help = true;
                            break;
                    }
                }
                else if (arg_params.Length == 2)
                {
                    switch (arg_params[0])
                    {
                        case "-port":
                            port = arg_params[1];
                            break;

                        case "-baudbps":
                            if (long.TryParse(arg_params[1], out baud_bps) == false)
                            {
                                Console.WriteLine("Invalid baud parameter: {0}", arg);
                                show_help = true;
                            }
                            break;

                        case "-ctbaddr":
                            if (byte.TryParse(arg_params[1], out ctb_read_start_addr) == false)
                            {
                                Console.WriteLine("Invalid Read Start Address parameter: {0}", arg);
                                show_help = true;
                            }
                            break;

                        case "-ctbreadlen":
                            if (byte.TryParse(arg_params[1], out ctb_read_len) == false)
                            {
                                Console.WriteLine("Invalid Read Length parameter: {0}", arg);
                                show_help = true;
                            }
                            break;

                        case "-test":
                            if (arg_params[1] == "base")
                            {
                                test_type = en_TestType.Base;
                            }
                            else if (arg_params[1] == "extended")
                            {
                                test_type = en_TestType.Extended;
                            }
                            else
                            {
                                test_type = en_TestType.Invalid;
                                Console.WriteLine("Invalid Test type: {0}", arg);
                                show_help = true;
                            }
                            break;

                        case "-dynids":
                            // split the parameter by commas and add them to the list
                            string[] s_dynids = arg_params[1].Split(',');
                            dyn_ids_to_query.Clear();

                            byte b_dyn_id;
                            foreach (string s_dynid in s_dynids)
                            {
                                if (byte.TryParse(s_dynid, out b_dyn_id))
                                {
                                    dyn_ids_to_query.Add(b_dyn_id);
                                }
                                else
                                {
                                    Console.WriteLine("Invalid ID specified in parameter: {0}, offending ID: '{1}'", arg_params[0], s_dynid);
                                    show_help = true;
                                }
                            }
                            break;

                        default:
                            Console.WriteLine("Invalid parameter: {0}", arg_params[0]);
                            show_help = true;
                            break;
                    }

                }
                else {
                    Console.WriteLine("Invalid parameter: {0}", arg);
                    show_help = true;
                    break;
                }
            }
        }
    }
}
